﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;

namespace SplitAndMerge
{
    public class ParserFunction
    {
        public ParserFunction()
        {
            m_impl = this;
        }

        // A "virtual" Constructor
        internal ParserFunction(ParsingScript script, string item, char ch, ref string action)
        {
            if (item.Length == 0 && (ch == Constants.START_ARG || !script.StillValid())) {
                // There is no function, just an expression in parentheses
                m_impl = s_idFunction;
                return;
            }

            m_impl = GetRegisteredAction(item, ref action);
            if (m_impl != null) {
                return;
            }

            m_impl = GetArrayFunction(item, script, action);
            if (m_impl != null) {
                return;
            }

            m_impl = GetFunction(item);
            if (m_impl != null) {
                return;
            }

            if (m_impl == s_strOrNumFunction && string.IsNullOrWhiteSpace(item)) {
                string problem = (!string.IsNullOrWhiteSpace(action) ? action : ch.ToString());
                string restData = ch.ToString() + script.Rest;
                throw new ArgumentException("Couldn't parse [" + problem + "] in " + restData + "...");
            }

            // Function not found, will try to parse this as a string in quotes or a number.
            s_strOrNumFunction.Item = item;
            m_impl = s_strOrNumFunction;
        }

        public static ParserFunction GetArrayFunction(string name, ParsingScript script, string action)
        {
            int arrayStart = name.IndexOf(Constants.START_ARRAY);
            if (arrayStart < 0) {
                return null;
            }

            string arrayName = name;

            int delta = 0;
            List<Variable> arrayIndices = Utils.GetArrayIndices(ref arrayName, ref delta);

            if (arrayIndices.Count == 0) {
                return null;
            }

            ParserFunction pf = ParserFunction.GetFunction(arrayName);
            GetVarFunction varFunc = pf as GetVarFunction;
            if (varFunc == null) {
                return null;
            }

            // we temporarily backtrack for the processing
            script.Backward(name.Length - arrayStart - 1);
            script.Backward(action != null ? action.Length : 0);
            // delta shows us how manxy chars we need to advance forward in GetVarFunction()
            delta -= arrayName.Length;
            delta += action != null ? action.Length : 0;

            varFunc.Indices = arrayIndices;
            varFunc.Delta = delta;
            return varFunc;
        }

        public static ParserFunction GetRegisteredAction(string name, ref string action)
        {
            ActionFunction actionFunction = GetAction(action);

            // If passed action exists and is registered we are done.
            if (actionFunction == null) {
                return null;
            }

            ActionFunction theAction = actionFunction.NewInstance() as ActionFunction;
            theAction.Name = name;
            theAction.Action = action;

            action = null;
            return theAction;
        }

        public static ParserFunction GetFunction(string item)
        {
            ParserFunction impl;
            // First search among local variables.

            if (s_locals.Count > 0) {
                Dictionary<string, ParserFunction> local = s_locals.Peek().Variables;
                if (local.TryGetValue(item, out impl)) {
                    // Local function exists (a local variable)
                    return impl;
                }
            }
            if (s_functions.TryGetValue(item, out impl)) {
                // Global function exists and is registered (e.g. pi, exp, or a variable)
                return impl.NewInstance();
            }

            return null;
        }

        public static ActionFunction GetAction(string action)
        {
            if (string.IsNullOrWhiteSpace(action)) {
                return null;
            }

            ActionFunction impl;
            if (s_actions.TryGetValue(action, out impl)) {
                // Action exists and is registered (e.g. =, +=, --, etc.)
                return impl;
            }

            return null;
        }

        public static bool FunctionExists(string item)
        {
            // If it is not defined locally, then check globally:
            return LocalNameExists(item) || GlobalNameExists(item);
        }

        public static void AddGlobalOrLocalVariable(string name, ParserFunction function)
        {
            function.Name = name;
            if (s_locals.Count > 0 && (LocalNameExists(name) || !GlobalNameExists(name))) {
                AddLocalVariable(function);
            } else {
                AddGlobal(name, function, false /* not native */);
            }
        }

        static bool LocalNameExists(string name)
        {
            if (s_locals.Count == 0) {
                return false;
            }
            var vars = s_locals.Peek().Variables;
            return vars.ContainsKey(name);
        }
        static bool GlobalNameExists(string name)
        {
            return s_functions.ContainsKey(name);
        }

        public static void RegisterFunction(string name, ParserFunction function,
                                            bool isNative = true)
        {
            AddGlobal(name, function, isNative);
        }

        public static void AddGlobal(string name, ParserFunction function,
                                     bool isNative = true)
        {
            function.isNative = isNative;
            s_functions[name] = function;

            if (string.IsNullOrWhiteSpace(function.Name)) {
                function.Name = name;
            }
            if (!isNative) {
                Translation.AddTempKeyword(name);
            }
            //Console.WriteLine("Registered function " + name);
        }

        public static void AddAction(string name, ActionFunction action)
        {
            s_actions[name] = action;
        }

        public static void AddLocalVariables(StackLevel locals)
        {
            s_locals.Push(locals);
        }

        public static void AddStackLevel(string name)
        {
            s_locals.Push(new StackLevel(name));
        }

        public static void AddLocalVariable(ParserFunction local)
        {
            local.m_isGlobal = false;
            StackLevel locals = null;
            if (s_locals.Count == 0) {
                locals = new StackLevel();
                s_locals.Push(locals);
            } else {
                locals = s_locals.Peek();
            }

            locals.Variables[local.Name] = local;
            Translation.AddTempKeyword(local.Name);
        }

        public static void PopLocalVariables()
        {
            s_locals.Pop();
        }

        public static int GetCurrentStackLevel()
        {
            return s_locals.Count;
        }

        public static void InvalidateStacksAfterLevel(int level)
        {
            while (s_locals.Count > level) {
                s_locals.Pop();
            }
        }

        public static void PopLocalVariable(string name)
        {
            if (s_locals.Count == 0) {
                return;
            }
            Dictionary<string, ParserFunction> locals = s_locals.Peek().Variables;
            locals.Remove(name);
        }

        public Variable GetValue(ParsingScript script)
        {
            return m_impl.Evaluate(script);
        }

        protected virtual Variable Evaluate(ParsingScript script)
        {
            // The real implementation will be in the derived classes.
            return new Variable();
        }

        // Derived classes may want to return a new instance in order to
        // not to use same object in calculations.
        public virtual ParserFunction NewInstance()
        {
            return this;
        }

        protected string m_name;
        public string Name { get { return m_name; } set { m_name = value; } }

        protected bool m_isGlobal = true;
        public bool isGlobal { get { return m_isGlobal; } set { m_isGlobal = value; } }

        protected bool m_isNative = true;
        public bool isNative { get { return m_isNative; } set { m_isNative = value; } }

        private ParserFunction m_impl;
        // Global functions and variables:
        private static Dictionary<string, ParserFunction> s_functions = new Dictionary<string, ParserFunction>();

        // Global variables:
        //private static Dictionary<string, ParserFunction> s_globals = new Dictionary<string, ParserFunction> ();

        // Global actions - function:
        private static Dictionary<string, ActionFunction> s_actions = new Dictionary<string, ActionFunction>();

        public class StackLevel
        {
            public StackLevel(string name = null)
            {
                Name = name;
                Variables = new Dictionary<string, ParserFunction>();
            }
            public string Name { get; set; }
            public Dictionary<string, ParserFunction> Variables { get; set; }
        }

        // Local variables:
        // Stack of the functions being executed:
        private static Stack<StackLevel> s_locals = new Stack<StackLevel>();
        public static Stack<StackLevel> ExecutionStack { get { return s_locals; } }

        private static StringOrNumberFunction s_strOrNumFunction =
          new StringOrNumberFunction();
        private static IdentityFunction s_idFunction =
          new IdentityFunction();
    }

    public abstract class ActionFunction : ParserFunction
    {
        protected string m_action;
        public string Action { set { m_action = value; } }
    }
}